#
#  Copyright (C) 2001, CodeStackers Ltd, See LICENCE for additional
#  licensing details
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; version 2.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
################################################################################
import Globals, AccessControl
from OFS.Folder import Folder
from BTrees.IOBTree import IOBTree
from BTrees.OOBTree import OOBTree
from Products.LocalFS.LocalFS import LocalFS
import time

# Valid password characters. Note that the letters and numbers are chosen so
# that they should not be confused from one another eg. l1I| can all look
# very similar in certain fonts
passchars = ['!', '#', '$', '%', '&', '+', '.', '2', '3', '4', '5', '6', '7',
             '8', '9', 'A', 'B', 'C', 'D', 'E', 'F', 'G', 'H', 'I', 'J', 'K',
             'L', 'M', 'N', 'P', 'Q', 'R', 'S', 'T', 'U', 'V', 'W', 'X', 'Y',
             'Z', 'b', 'c', 'd', 'e', 'f', 'g', 'h', 'i', 'j', 'k', 'm', 'n',
             'o', 'p', 'q', 'r', 's', 't', 'u', 'v', 'w', 'x', 'y', 'z']
def gen_password(password_length=8):
    import random
    password = ''
    for i in xrange(password_length):
        password += random.choice(passchars)
    return password

class CategoryError(Exception):
    pass

class GroupError(Exception):
    pass

class StatusError(Exception):
    pass

class ObserverError(Exception):
    pass

class UserError(Exception):
    pass

class MailError(Exception):
    pass

class PriorityError(Exception):
    pass

class IssueComment(Globals.Persistent):
    __allow_access_to_unprotected_subobjects__ = 1

    def __init__(self, comment, user_id):
        self.submit_time = int(time.time())
        self.comment = comment
        self.submitted_by = user_id

class IssueAttachment(IssueComment):
    __allow_access_to_unprotected_subobjects__ = 1

    def __init__(self, attachment_id, comment, user_id):
        IssueComment.__init__(self, comment, user_id)
        self.attachment_id = attachment_id 

class Issue(Globals.Persistent):
    __allow_access_to_unprotected_subobjects__ = 1
    
    meta_type = 'Issue'

    def __init__(self, IssueTracker, user_id, summary, id):
        self.issuetrack = IssueTracker
        
        # initialise our BTrees
        self.comments = IOBTree()
        self.comments_length = 0
        
        self.audit_trail = IOBTree()
        self.audit_trail_length = 0
        
        self.attachments = IOBTree()
        self.attachments_length = 0
        
        self.submit_time = int(time.time())
        
        self.set_submitted_by(user_id, user_id)
        self.set_summary(user_id, summary)

        self.id = id

        self.priority = 1
        
        self.status = ''
        self.category = None
        self.group = None
        self.assigned_to = 'nobody'

        self.observers = []

        self.last_notification_time = 0

    def add_observer(self, c_user_id, user_id):
        if user_id in self.observers:
            raise ObserverError, '%s is already an observer' % user_id
        
        self.__add_audit_trail(c_user_id, 'add observer', None, user_id)

        self.observers.append(user_id)
        
        self._p = 1
        
        return len(self.observers)

    def remove_observer(self, c_user_id, user_id):
        if user_id not in self.observers:
            raise ObserverError, '%s is not an observer' % user_id

        self.__add_audit_trail(c_user_id, 'remove observer', None, user_id)
        self.observers.remove(user_id)
        
        self._p = 1
        
        return len(self.observers)

    def is_observer(self, user_id):
        user_id = str(user_id.strip())
        if user_id in self.observers:
            return 1
        else:
            return 0
    
    def __add_audit_trail(self, user_id, comment, before, after):
        """ Add a new event to the audit trail

        """
        a_time = int(time.time())
        # add a list to the audit trail to avoid a race condition
        # wiping out the last entry
        event = [(user_id, comment, before, after)]
        if self.audit_trail.has_key(a_time):
            self.audit_trail[a_time] += event
        else:
            self.audit_trail[a_time] = event
        self.audit_trail_length += 1

        self._p = 1

    def set_category(self, c_user_id, category):
        category = category.strip()
        # make sure that we have a nice clean list
        categories = [test.strip() for test in self.issuetrack.categories]
        if category not in categories:
            raise CategoryError, 'category "%s" does not exist' % category
        
        self.__add_audit_trail(c_user_id, 'set category',
                               self.category, category)
        self.category = category

    def set_submitted_by(self, c_user_id, submitted_by):
        # submitted_by is _not_ a string, it is wrapped in an Aquisition wrapper
        submitted_by = str(submitted_by).strip()
        visible_users =  [test.strip() for test in \
                          self.issuetrack.visible_users]
        if submitted_by not in visible_users:
            raise UserError, 'user "%s" is not a valid user for submitted_by' %\
                  submitted_by
        
        self.__add_audit_trail(c_user_id, 'set submitted_by',
                               getattr(self, 'submitted_by', None),
                               submitted_by)
        self.submitted_by = str(submitted_by)
        
    def set_group(self, c_user_id, group):
        group = group.strip()
        groups = [test.strip() for test in self.issuetrack.groups]
        if group not in groups:
            raise GroupError, 'group "%s" does not exist' % group
        
        self.__add_audit_trail(c_user_id, 'set group', self.group, group)
        self.group = group

    def set_status(self, c_user_id, status):
        status = status.strip()
        statuses = [test.strip() for test in self.issuetrack.statuses]
        if status not in statuses:
            raise GroupError, 'status "%s" does not exist' % status
        
        self.__add_audit_trail(c_user_id, 'set status', self.status, status)
        self.status = status
        
    def set_priority(self, c_user_id, priority):
        priority = int(priority)
        if not (priority >= 1 and priority <= self.issuetrack.max_priority):
            raise PriorityError, 'priority "%s" is not a valid priority' % \
                  priority
        
        self.__add_audit_trail(c_user_id, 'set priority',
                               self.priority, priority)
        self.priority = priority
        
    def set_summary(self, c_user_id, summary):
        summary = summary.strip()
        
        self.__add_audit_trail(c_user_id, 'set summary',
                               getattr(self, 'summary', None),
                               summary)
        self.summary = summary
        
    def set_assignment(self, c_user_id, assigned_to):
        assigned_to = assigned_to.strip()
        visible_users = [test.strip() for test in self.issuetrack.visible_users]
        if assigned_to not in visible_users:
            raise UserError, 'user "%s" is not a valid user for assignment' % \
                  assigned_to
        
        self.__add_audit_trail(c_user_id, 'set assignment',
                               self.assigned_to, assigned_to)
        self.assigned_to = assigned_to
        
    def add_comment(self, user_id, comment):
        comment = comment.strip()
        
        self.__add_audit_trail(user_id, 'add comment', None, comment)
        comment_obj = IssueComment(comment, user_id)
        
        if self.comments.has_key(comment_obj.submit_time):
            self.comments[comment_obj.submit_time] += [comment_obj]
        else:
            self.comments[comment_obj.submit_time] = [comment_obj]
        self.comments_length += 1
        
        self._p = 1

    def add_attachment(self, user_id, filename, comment):
        filename = filename.split('\\')[-1]
        filename = filename.split('/')[-1]
        
        attachment_id = str(int(time.time()))+'_'+filename.strip()
        
        comment=comment.strip()
        
        self.__add_audit_trail(user_id, 'add attachment', None, comment)
        attachment_obj = IssueAttachment(attachment_id, comment, user_id)
        
        if self.attachments.has_key(attachment_obj.submit_time):
            self.attachments[attachment_obj.submit_time] += [attachment_obj]
        else:
            self.attachments[attachment_obj.submit_time] = [attachment_obj]
        self.attachments_length += 1
        
        self._p = 1
        
        return attachment_id
    
    def get_last_change_time(self):
        return self.audit_trail.maxKey()
    
    last_change_time = get_last_change_time
    
    def get_comments(self):
        return self.comments.values()
    
    def get_attachments(self):
        return self.attachments.values()
    
    def get_attachment(self, attachment_id):
        if self.attachments.has_key(attachment_id):
            return self.attachments[attachment_id]
        else:
            return None
        
    def get_number_comments(self):
        return self.comments_length
    
    def get_number_attachments(self):
        return self.attachments_length

class User(Folder):
    __allow_access_to_unprotected_subobjects__ = 1
    
    default_email = 'my@email.address'

    password_length = 10
    password_subject = """ZShrink Sign-on"""
    password_message = """
    You requested that your ZShrink password be sent to you, for username
    %(user_id)s.

    Your password is: %(password)s
    """
    security = AccessControl.ClassSecurityInfo()
    security.setPermissionDefault('View', ('Manager', 'Owner'))
    
    def __init__(self, user_id):
        self.manage_addProperty('user_id',
                                 user_id, 'string')
        self.manage_addProperty('email',
                                 self.default_email, 'string')
        self.manage_addProperty('email_own_bugs',
                                 0, 'int')
        
    security.declarePrivate('_set_email')
    def _set_email(self, email):
        self.email = email.strip()

    security.declareProtected('View', 'get_email')
    def get_email(self):
        return self.email or self.default_email

    security.declarePrivate('_is_email_set')
    def _is_email_set(self):
        if self.email != self.default_email:
            return 1
        else:
            return 0

    security.declarePrivate('_send_simple_email')
    def _send_simple_email(self, subject, message):
        if not self._is_email_set:
            raise MailError, \
                  'Attempted to send email, but no email address is set'
         
        mailhost = self._get_mailhost()
        
        if not mailhost:
            raise MailError, \
                  'Attempted to send email, but no MailHost set'
        
        mailhost.simple_send(self.get_email(), self.email_from_addr,
                             subject, message)

    security.declarePrivate('_get_user')
    def _get_user(self, acl_users_only=0):
        """ Get the user object from acl_users

            if acl_user_only is set, we only retrieve the acl_users object, not
            the actual user.
        
        """
        curr = self
        while 1:
            acl_users = getattr(curr, 'acl_users', None)
            if not acl_users:
                return None
            user = acl_users.getUser(self.user_id)
            if user:
                if acl_users_only:
                    return acl_users
                return user
            if acl_users._isTop():
                return None
            curr = curr.aq_parent

    security.declarePrivate('_send_password')
    def _send_password(self, password):
        """ Send a password to the users email address

        """
        self._send_simple_email(self.password_subject,
                                self.password_message % \
                                {'user_id': self.user_id,
                                 'password': password})
        
    security.declarePrivate('_generate_password')
    def _generate_password(self):
        """ Generate a random password

        """
        password = gen_password(self.password_length)
        
        return password

    security.declarePrivate('_get_mailhost')
    def _get_mailhost(self):
        mailhost_id = getattr(self, 'mailhost_id', 'MailHost')
        mailhost = getattr(self, mailhost_id, None)
        
        return mailhost

    security.declareProtected('View', 'insert_user')
    def insert_user(self, password=None, mail_password=1):
        """ Insert the user into acl_users.

            If we are not given a password, generate a new password and send
            it to the users email address.
            
        """
        if self._get_user():
            raise UserError, 'User %s already exists, refusing to insert' \
                  % self.user_id
        
        if not password:
            password = self._generate_password()
            
        password = password.strip()
        
        self.acl_users._doAddUser(self.user_id, password, [], []) 
        
        self._send_password(password)
        
    security.declareProtected('View', '_set_password')
    def _set_password(self, password, confirm):
        """ Set the users password in acl_users.

        """
        user = self._get_user()
        if not user:
            raise UserError, \
                  'User %s does not exist, unable to set their password'
        password = password.strip()
        confirm = confirm.strip()
        acl_users = self._get_user(acl_users_only=1)
        if acl_users and password != self.get_password() \
           and password == confirm:
            acl_users._doChangeUser(self.user_id, password,
                                    user.getRoles(), user.getDomains())
    
    security.declareProtected('View', 'get_password')
    def get_password(self):
        user = self._get_user()

        if user == None:
            return ''
            
        return user._getPassword()
        
    security.declareProtected('View', 'set_email_own_bugs')
    def set_email_own_bugs(self, state):
        self.set_email_own_bugs = not(not(state)) # state should be 1 or 0


    #
    # Methods related to the notification framework
    ############################################################################
    security.declareProtected('View', 'notify_local')
    def notify_local(self, issue):
        """ Called to notify an observer of changes to an individual issue.

            Override to provide customised notification
            
        """
        import sys
        if hasattr(self, 'my_notify_local'):
            getattr(self, 'my_notify_local')(issue=issue)
            return 2
        
        if self.get_email() == self.default_email:
            return 0
        
        subject = '%s: Changes to Issue #%s' % (self.get_zshrink_id(), issue.id)
        body = 'Issue #%(id)s has had changes made -- \n\n' \
               '%(url)s/edit_issue_screen?issue_id=%(id)s\n\n' \
               % {'id':issue.id, 'url':self.get_absolute_url()}

        body_h = 'modified'.ljust(23)+\
                 'old value'.ljust(23)+\
                 'new value'.ljust(23)+'\n'
        body_h += '-'*69
        
        comments = []
        attachments = []

        do_h = 0
        for key in issue.audit_trail.keys(int(issue.last_notification_time)+1):
            changes = issue.audit_trail[int(key)]
            for change in changes:
                modified = change[1] or '-'
                if modified == 'add comment':
                    new_value = str(change[3]) or '-'
                    comments.append(new_value)
                elif modified == 'add attachment':
                    new_value = str(change[3]) or '-'
                    attachments.append(new_value)
                else:
                    do_h = 1
                    modified = str(modified).strip()[:22].ljust(23)
                    old_value = change[2] or '-'
                    old_value = str(old_value).strip()[:22].ljust(23)
                    new_value = change[3] or '-'
                    new_value = str(new_value).strip()[:22].ljust(23)
                    body_h += '\n%s%s%s' % (modified, old_value, new_value)
                    
        if do_h:
            body += body_h
            
        body += '\n\n'
        
        if comments:
            do_h = 1
            body += 'new comment\n'
            body += '-----------\n\n'
            for comment in comments:
                body += '+++\n'
                body += comment
                body += '\n+++\n\n'
                
        if attachments:
            do_h = 1
            body += 'new attachment\n'
            body += '--------------\n\n'
            for attachment in attachments:
                body += '+++\n'
                body += attachment
                body += '\n+++\n\n'
                
        if do_h:
            self._send_simple_email(subject, body)
        
        return 1
    
    security.declareProtected('View', 'notify_global')
    def notify_global(self, issue):
        """ Called to notify an observer of changes to any issue.

            Override to provide customised notification
            
        """
        import sys
        if hasattr(self, 'my_notify_global'):
            getattr(self, 'my_notify_global')(issue=issue)
            return 2

        # we can't do much if the user hasn't set their email address
        if self.get_email() == self.default_email:
            return 0
        
        subject = '%s: Changes to Issue #%s' % (self.get_zshrink_id(), issue.id)
        body = 'Issue #%(id)s has had changes made -- \n\n' \
               '%(url)s/edit_issue_screen?issue_id=%(id)s' \
               % {'id':issue.id, 'url':self.get_absolute_url()}
        
        self._send_simple_email(subject, body)        
        
        return 1
        
class IssueTrack(Folder):
    from Products.PageTemplates.ZopePageTemplate import manage_addPageTemplate
    from Products.LocalFS.LocalFS import manage_addLocalFS
    from Products.ZCatalog.ZCatalog import manage_addZCatalog
    from OFS.Folder import manage_addFolder
    from OFS.DTMLMethod import addDTMLMethod
    from OFS.Image import manage_addImage
    from Products.MailHost.MailHost import manage_addMailHost
    
    meta_type = 'ZShrink'
    
    security = AccessControl.ClassSecurityInfo()
    security.declareProtected('View management screens', 'manage_workspace',
                              'manage_main',
                              'manage')
    
    default_statuses = ('open', 'closed', 'deleted', 'awaiting triage',)
    default_auto_open_status = 'open'
    default_initial_status = 'awaiting triage'
    
    default_groups = ('feature request', 'not a bug', 'works for me', 'bug')
    
    default_categories = ('general',)
    
    default_max_priority = 9
    
    default_max_display_num = 10
    
    default_suppress_statuses = ('deleted',)
    
    default_global_observers = ('',)
    
    default_email_from_addr = 'zshrink@codestackers.com'

    default_mailhost_id = 'MailHost'
    
    _times = (('1min', 60), ('10min', 600), ('30min', 1800), ('1h', 3600),
              ('3h', 10800), ('6h', 21600), ('12h', 43200), ('1d', 86400),
              ('2d', 172800), ('3d', 259200), ('1w', 604800), ('2w', 1209600),
              ('1m', 2419200), ('3m', 7257600), ('6m', 14515200),
              ('1y', 29030400))
    
    user_folder_name = 'Users'
    
    # the current version of the software -- NOTE: there is also the current
    # version of the software instantiated in the database '_version'
    version = 0.42
    
    def __init__(self, id):
        self.id = id
        
        self.issues = IOBTree()
        self.issues_length = 0
        
        self._version = self.version
        
        self.next_id = 1
        
    #
    # Methods related to the issues database
    ############################################################################
    security.declarePrivate('__add_to_issues')
    def __add_to_issues(self, key, item):
        self.issues[key] = item
        self.issues_length += 1
        
    security.declarePrivate('__del_from_issues')
    def __del_from_issues(self, key):
        del(self.issues[key])
        self.issues_length -= 1
        
    security.declareProtected('View', 'submit_issue')
    def submit_issue(self, summary, category, group, comment=None,
                           file_attachment=None, file_description=None,
                           notify_state=None,
                           REQUEST={}, RESPONSE=None):
        """ Submit a new issue to the tracker.
        
        """
        user_id = str(getattr(REQUEST, 'AUTHENTICATED_USER', 'nobody'))
        
        issue = Issue(self, user_id, summary, self.next_id)
        
        if comment:
            issue.add_comment(user_id, comment)
            
        if file_attachment and file_attachment.filename:
            attachment_id = issue.add_attachment(user_id,
                                                 file_attachment.filename,
                                                 file_description)
            self.file_attachments.manage_upload(file_attachment, attachment_id)
            
        # set the initial priority to half the maximum priority
        issue.set_priority(user_id, self.max_priority/2)
        
        issue.set_status(user_id, self.initial_status)
        
        if group:
            issue.set_group(user_id, group)
            
        if category:
            issue.set_category(user_id, category)
            
        self.__add_to_issues(self.next_id, issue)
        
        self.next_id += 1
        
        self.Catalog.catalog_object(issue, str(issue.id))
        
        if notify_state:
            try:
                issue.add_observer(user_id, user_id)
            except ObserverError:
                # this should NEVER happen, since the issue is new!
                raise
            
        self._notify_observers(issue)
        
        if RESPONSE:
            return RESPONSE.redirect('issue_summary_screen')
        
    security.declareProtected('View', 'modify_issue')
    def modify_issue(self, issue_id, summary, priority,
                           category, group,
                           comment='',
                           file_attachment=None,
                           file_description=None,
                           status=None,
                           assigned_to=None,
                           notify_state=None,
                           REQUEST={}, RESPONSE=None):
        """ Modify an issue that already exists in the database.
        
        """
        user_id = str(getattr(REQUEST, 'AUTHENTICATED_USER', 'nobody'))
        issue = self.get_issue(issue_id)
        
        if file_attachment and file_attachment.filename:
            attachment_id = issue.add_attachment(user_id,
                                                 file_attachment.filename,
                                                 file_description)
            self.file_attachments.manage_upload(file_attachment, attachment_id)
            
        comment = comment.strip()
        change_status = 0
        
        if comment:
            issue.add_comment(user_id, comment)
            
        if summary and issue.summary != summary:
            issue.set_summary(user_id, summary)
            
        if priority and issue.priority != int(priority):
            issue.set_priority(user_id, int(priority))
            
        if group and issue.group != group:
            issue.set_group(user_id, group)
            
        if category and issue.category != category:
            issue.set_category(user_id, category)
            
        if assigned_to and issue.assigned_to != assigned_to:
            issue.set_assignment(user_id, assigned_to)
            change_status = 1
            
        if status and issue.status != status:
            issue.set_status(user_id, status)
        elif change_status:
            statuses = [test.strip() for test in self.statuses]
            if self.auto_open_status in statuses:
                issue.set_status(user_id, self.auto_open_status)
                
        if notify_state != None:
            if notify_state:
                try:
                    issue.add_observer(user_id, user_id)
                except ObserverError:
                    # if we are already an observer then, well, good. 
                    pass
            else:
                try:
                    issue.remove_observer(user_id, user_id)
                except ObserverError:
                    # if we aren't an observer, it doesn't really matter, since
                    # we wanted to stop being an observer anyway :)
                    pass
                
        self.Catalog.uncatalog_object(str(issue.id))
        self.Catalog.catalog_object(issue, str(issue.id))
        
        self._notify_observers(issue)
        
        if RESPONSE:
            return RESPONSE.redirect('issue_summary_screen')
        
    security.declareProtected('View', 'get_issues')
    def get_issues(self, REQUEST={}):
        """ Get all issues from the database, based on a given criteria.
        
        """
        import time
        sort_order = str(REQUEST.get('sort_order', 'ascending')).strip()
        sort_on = str(REQUEST.get('sort_on', 'id')).strip()
        start_val = int(REQUEST.get('start_val', 1)) - 1
        end_val = start_val + (self.max_display_num)
        
        suppress_statuses = [test.strip() for test in self.suppress_statuses]
        unsuppress_statuses = []
        
        status = REQUEST.get('status', ['open'])
        status = [test.strip() for test in status]
        
        for item in suppress_statuses:
            if item in status:
                unsuppress_statuses.append(item)
                
        if 'any' in tuple(status):
            status = ''
            
        submitted_by = REQUEST.get('submitted_by', ['any'])
        submitted_by = [test.strip() for test in submitted_by]
        if 'any' in submitted_by:
            submitted_by = ''
            
        assigned_to = REQUEST.get('assigned_to', ['any'])
        assigned_to = [test.strip() for test in assigned_to]
        if 'any' in assigned_to:
            assigned_to = ''
            
        group = REQUEST.get('group', ['any'])
        group = [test.strip() for test in group]
        if 'any' in group:
            group = ''
            
        category = REQUEST.get('category', ['any'])
        category = [test.strip() for test in category]
        if 'any' in category:
            category = ''
            
        id = str(REQUEST.get('id', '')).strip()
        if id:
            id = int(id)
            
        priority = str(REQUEST.get('priority', '')).strip()
        if priority:
            priority = int(priority)
            
        summary = str(REQUEST.get('summary', '')).strip()
        if summary:
            summary = summary.split(' ')
            
        sub_time = str(REQUEST.get('sub_time', '')).strip()
        if sub_time:
            for word_time, int_time in self._times:
                if word_time == sub_time:
                    sub_time = int(time.time())-int_time
                    break
                
        lc_time = str(REQUEST.get('lc_time', '')).strip()
        if lc_time:
            for word_time, int_time in self._times:
                if word_time == lc_time:
                    lc_time = int(time.time())-int_time
                    break

        results = self.Catalog.searchResults(
            {'id': id,
             'priority': priority,
             'priority_usage': 'range:min',
             'status': status,
             'submitted_by': submitted_by,
             'group': group,
             'category': category,
             'submit_time': sub_time,
             'submit_time_usage': 'range:min',
             'last_change_time': lc_time,
             'last_change_time_usage':'range:min',
             'assigned_to': assigned_to,
             'summary': summary,
             'summary_usage': 'query',
             'sort_on': sort_on,
             'sort_order': sort_order})
        
        if not results:
            return [], 0, 0

        orig_length = len(results)
        length = orig_length

        results_set = []
        index = 0; pos = 0
        import time
        a = time.time()
        # sort the results set, removing deleted items where necessary.
        # this has got to be non-optimal.
        while 1:
            try:
                result = results[index]
            except IndexError:
                break
            if result.status in suppress_statuses and \
               result.status not in unsuppress_statuses:
                length -= 1
            elif pos >= start_val and pos < end_val:
                results_set.append(result)
                pos += 1
            else:
                pos += 1

            index += 1
        b = time.time()
        print b-a
        return results_set, length, orig_length-length
    
    security.declareProtected('View', 'get_issue')
    def get_issue(self, issue_id):
        """ Get an issue from the database, based on its issue_id.
        
        """
        if self.issues.has_key(int(issue_id)):
            return self.issues[int(issue_id)]
        else:
            return None
        
    security.declareProtected('View', 'count_issues')
    def count_issues(self):
        """ Count the number of issues in the database.

        """
        return self.issues_length
    
    security.declareProtected('View management screens', 'export_issues_to_xml')
    def export_issues_to_xml(self, RESPONSE):
        """ Export all the issues in the database as an XML stream

            TODO: Update so that it is actually useful!
        """
        import urllib
        
        RESPONSE.setHeader('Content-Type', 'application/xml')
        RESPONSE.write("<issues>\n")
        
        for issue in self.issues.values():
            audit_trail = issue.audit_trail
            
            RESPONSE.write(('  <issue id="%s" priority="%s" status="%s" '
                            'category="%s" group="%s" assigned_to="%s" '
                            'submit_time="%s" submitted_by="%s" '
                            'summary="%s">\n' %
                            (issue.id, issue.priority, issue.status,
                             issue.category, issue.group, issue.assigned_to,
                             issue.submit_time, issue.submitted_by,
                             urllib.quote_plus(issue.summary))
                            ))

            for comment_list in issue.comments.values():
                for comment in comment_list:
                    RESPONSE.write(('    <comment submit_time="%s" '
                                    'submitted_by="%s">\n' %
                                   (issue.submit_time, issue.submitted_by)
                                    ))
                    RESPONSE.write('%s' % urllib.quote_plus(comment.comment))
                    RESPONSE.write('    </comment>\n')
                    
            for attachment_list in issue.attachments.values():
                for attachment in attachment_list:
                    RESPONSE.write(('    <attachment submit_time="%s" '
                                    'submitted_by="%s" '
                                    'attachment_id="%s">\n' %
                                   (issue.submit_time, issue.submitted_by,
                                    issue.attachment_id)
                                    ))
                    RESPONSE.write('%s' % urllib.quote_plus(comment.comment))
                    RESPONSE.write('    </attachment>\n')
                    
            RESPONSE.write("  </issue>\n")
                
        RESPONSE.write("</issues>")
        RESPONSE.flush()
        
    #
    # Methods related to the notification framework
    ############################################################################
    security.declarePrivate('add_global_observer')
    def add_global_observer(self, user_id):
        self.global_observers.append(user_id)
        if user_id in self.global_observers:
            raise ObserverError, '%s is already an observer' % user_id
        
        self.global_observers.append(user_id)
        
        self._p = 1
        
        return len(self.global_observers)

    security.declarePrivate('remove_global_observer')
    def remove_global_observer(self, user_id):
        self.global_observers.append(user_id)
        if user_id not in self.global_observers:
            raise ObserverError, '%s is not an observer' % user_id
        
        self.global_observers.append(user_id)
        
        self._p = 1
        
        return len(self.global_observers)

    security.declarePrivate('_notify_observers')
    def _notify_observers(self, issue):
        # notify local observers
        for observer in issue.observers:
            try:
                self.get_user(observer).notify_local(issue)
            except Exception, x:
                print 'caught exception %s' % x
                issue.observers.remove(observer)
        # notify global observers
        for observer in self.global_observers:
            try:
                self.get_user(observer).notify_global(issue)
            except Exception, x:
                print 'caught exception %s' % x
                self.global_observers.remove(observer)
                
        issue.last_notification_time = int(time.time())

    #
    # User management methods
    ############################################################################
    security.declareProtected('View', 'check_user')
    def check_user(self, user_id):
        """ Check the existance of a user

        """
        user_id = str(user_id)
        user_folder = getattr(self, self.user_folder_name)

        return hasattr(user_folder, user_id)

    security.declareProtected('View', 'get_user')
    def get_user(self, user_id):
        """ Get a user object from the database, based on its user_id
        
        """
        user_id = str(user_id)
        user_folder = getattr(self, self.user_folder_name)
        
        return getattr(user_folder, user_id, None)

    security.declareProtected('View', 'add_user')
    def add_user(self, user_id, email, REQUEST={}, RESPONSE=None):
        """ Add a user that exists in acl_users to the system.
        
        """
        user = User(user_id)
        user._set_email(email)

        user_folder = getattr(self, self.user_folder_name)
        
        user_folder._setObject(user_id, user, user_id)

        self.visible_users.append(user_id)
        self._p = 1
        
        if RESPONSE:
            RESPONSE.redirect('')

    security.declarePublic('add_user_gen_pass')
    def add_user_gen_pass(self, user_id, email, REQUEST={}, RESPONSE=None):
        """ Add a new user, including having the password emailed to the user.
        
        """
        user = User(user_id)
        user._set_email(email)
        
        user_folder = getattr(self, self.user_folder_name)
        
        user_folder._setObject(user_id, user, user_id)
        
        user = getattr(user_folder, user_id)
        
        try:
            user.insert_user(mail_password=1)
        except:
            user_folder._delObject(user_id)
            raise
        
        self.visible_users.append(user_id)
        self._p = 1
        
        if RESPONSE:
            RESPONSE.redirect('')
            
    security.declareProtected('View', 'modify_own_user')
    def modify_own_user(self, email=None, password=None, confirm=None,
                        REQUEST={}, RESPONSE=None):
        """ Modify a user object, with more protection for less priviledged
        users.
        
        """
        user_id = str(getattr(REQUEST, 'AUTHENTICATED_USER', 'nobody'))
        
        self.modify_user(user_id, email, password, confirm, REQUEST, RESPONSE)
        
    security.declareProtected('View management screens', 'modify_user')
    def modify_user(self, user_id, email=None, password=None, confirm=None,
                    REQUEST={}, RESPONSE=None):
        """ Modify a user object
        
        """
        users_folder = getattr(self, self.user_folder_name)
        if hasattr(users_folder, user_id):
            user = getattr(users_folder, user_id)
            
            if email:
                user._set_email(email)
                
            if password and confirm:
                user._set_password(password, confirm)
                
        if RESPONSE:
            return RESPONSE.redirect('issue_summary_screen')
        
    security.declareProtected('View', 'get_visible_users')
    def get_visible_users(self):
        """ Return visible users, sorted and stripped.
        
        """
        visible_users = [test.strip() for test in self.visible_users]
        visible_users.sort()
        
        return visible_users
    
    #
    # Highly dangerous methods that should be very rarely used!
    ############################################################################
    security.declareProtected('View management screens', 'remove_all_issues')
    def remove_all_issues(self):
        """ Remove all issues from the database, and reinitialise the tree
        
        """
        del(self.issues)
        self.issues = OOBTree()
        self.issues_length = 0
        self.update_catalog()
        
    #
    # Assorted helper methods
    ############################################################################
    security.declarePublic('url_list')
    def url_list(self, name, alist):
        """ Take a list and make it into a valid portion of a HTTP GET string
        
            NOTE: I'm almost certain that this duplicates functionality in the
            TAL Utils.
            
        """
        import string, urllib
        
        result = []
        for item in alist:
            result.append('%s:list=%s' % (name, urllib.quote(item)))

        return string.join(result, '&')
    
    security.declarePublic('calc_next')
    def calc_next(self, num):
        """ Calculate the next display start point.

        """
        return int(num)+self.max_display_num

    security.declarePublic('calc_prev')
    def calc_prev(self, num):
        """ Calculate the previous display start point.
        
        """
        return int(num)-self.max_display_num

    
    security.declarePublic('get_shrink_id')
    def get_zshrink_id(self):
        """ Return the id of this ZShrink when instantiated.
        
        """
        return self.id

    security.declarePublic('get_absolute_url')
    def get_absolute_url(self):
        """ Return the absolute URL of the ZShrink instance.

        """
        return self.absolute_url()
    
    #
    # Product initialisation methods
    ############################################################################
    security.declarePrivate('__read_template')
    def __read_template(self, relative_path):
        import os
        
        TEMPLATE_PATH = os.path.join(Globals.package_home(globals()),
                                     relative_path)
        
        f = open(TEMPLATE_PATH)

        return f.read()    

    security.declarePrivate('__add_page_templates')
    def __add_page_templates(self, templates):
        for id, title, rel_path in templates:
            self.manage_addPageTemplate(id, title,
                                        self.__read_template(rel_path))

    security.declarePrivate('__add_dtml_methods')
    def __add_dtml_methods(self, methods):
        for id, title, rel_path in methods:
            self.manage_addDTMLMethod(id, title,
                                      self.__read_template(rel_path))

    security.declarePrivate('__add_images')
    def __add_images(self, images):
        import StringIO
        for id, title, rel_path in images:
            f = StringIO.StringIO(self.__read_template(rel_path))
            f.filename = id
            self.manage_addImage(id, f, title)
            
    security.declareProtected('View management screens', 'set_mailhost') 
    def set_mailhost(self, mailhost_id):
        if hasattr(self, mailhost_id):
            self._updateProperty('mailhost_id', mailhost_id)
        else:
            self._updateProperty('mailhost_id', '')

    security.declarePrivate('post_install')
    def post_install(self, file_attachment_path):
        self.__initialise_product(file_attachment_path)

    security.declareProtected('Apply fixes', 'update_user_interface')
    def update_user_interface(self):
        """ Update the User Interface from the file system.
        
        """
        try: self._delObject('layout_template')
        except AttributeError: pass
        try: self._delObject('edit_issue_screen')
        except AttributeError: pass
        try: self._delObject('issue_summary_screen')
        except AttributeError: pass
        try: self._delObject('submit_issue_screen')
        except AttributeError: pass
        try: self._delObject('index_html')
        except AttributeError: pass
        try: self._delObject('navigation_bar_top')
        except AttributeError: pass
        try: self._delObject('navigation_bar_bottom')
        except AttributeError: pass
        try: self._delObject('filter_widget')
        except AttributeError: pass
        try: self._delObject('issues_widget')
        except AttributeError: pass
        try: self._delObject('modify_user_screen')
        except AttributeError: pass
        try: self._delObject('add_user_screen')
        except AttributeError: pass
        try: self._delObject('add_existing_user_screen')
        except AttributeError: pass
                                     
        self.__add_page_templates([
            ('layout_template',
             'Layout Template',
             'templates/default/layout_template.zpt'),
            ('edit_issue_screen',
             'Issue Editing Screen',
             'templates/default/edit_issue_screen.zpt'),
            ('issue_summary_screen',
             'Issue Summary Screen',
             'templates/default/issue_summary_screen.zpt'),
            ('submit_issue_screen',
             'Submit Issue Screen',
             'templates/default/submit_issue_screen.zpt'),
            ('index_html',
             'Default Screen',
             'templates/default/index_html.zpt'),
            ('navigation_bar_top',
             'Top Navigation Bar Template',
             'templates/default/navigation_bar_top.zpt'),
            ('navigation_bar_bottom',
             'Bottom Navigation Bar Template',
             'templates/default/navigation_bar_bottom.zpt'),
            ('filter_widget',
             'Filter Widget Template',
             'templates/default/filter_widget.zpt'),
            ('issues_widget',
             'Issues Widget Template',
             'templates/default/issues_widget.zpt'),
            ('modify_user_screen',
             'Modify User Screen',
             'templates/default/modify_user_screen.zpt'),
            ('add_user_screen',
             'Add User Screen',
             'templates/default/add_user_screen.zpt'),
            ('add_existing_user_screen',
             'Add Existing User Screen',
             'templates/default/add_existing_user_screen.zpt'),
            ])
                                 
        try:
            self.manage_delObjects(['default.css'])
        except 'BadRequest':
            pass
                                   
        self.__add_dtml_methods([('default.css',
                                  'Default Cascading Style Sheet',
                                  'templates/default/default.css')])
        try:
            self.manage_delObjects(['codestackers_logo.gif'])
        except 'BadRequest':
            pass

        self.__add_images([('codestackers_logo.gif',
                            'CodeStackers Logo',
                            'templates/default/codestackers_logo.gif')])

    security.declareProtected('View management screens', 'update_catalog')
    def update_catalog(self):
        """ Update the catalog from scratch. Mostly just for upgrading.
        
        """
        try:
            self.manage_delObjects(['Catalog'])
        except 'BadRequest':
            pass

        self.manage_addZCatalog('Catalog', 'Issue Tracker Catalog')

        for column in ('priority', 'submit_time', 'last_change_time', 'status',
                       'category', 'group', 'submitted_by', 'assigned_to'):
            self.Catalog.manage_addColumn(column)

        for index in ('priority', 'submit_time', 'last_change_time', 'status',
                      'category', 'group', 'submitted_by', 'assigned_to'):
            self.Catalog.manage_addIndex(index, 'FieldIndex')
            
        for index in ('summary',):
            self.Catalog.manage_addIndex(index, 'TextIndex')

        for issue_key in self.issues.keys():
            issue = self.issues[issue_key]
            self.Catalog.catalog_object(issue, str(issue.id))

    security.declareProtected('View management screens', 'update_properties')
    def update_properties(self):
        """ Update the property sheet from scratch. Mostly just for upgrading

        """
        # get a copy of our old properties
        default_statuses = getattr(self, 'statuses', self.default_statuses)
        default_auto_open_status = getattr(self, 'auto_open_status',
                                           self.default_auto_open_status)
        default_initial_status = getattr(self, 'initial_status',
                                         self.default_initial_status)
        default_categories = getattr(self, 'categories',
                                     self.default_categories)
        default_groups = getattr(self, 'groups', self.default_groups)
        default_max_priority = getattr(self, 'max_priority',
                                       self.default_max_priority)
        default_max_display_num = getattr(self, 'max_display_num',
                                          self.default_max_display_num)
        default_suppress_statuses = getattr(self, 'suppress_statuses',
                                            self.default_suppress_statuses)
        default_global_observers = getattr(self, 'global_observers',
                                           self.default_global_observers)
        default_email_from_addr = getattr(self, 'email_from_addr',
                                          self.default_email_from_addr)
        default_mailhost_id = getattr(self, 'mailhost_id',
                                     self.default_mailhost_id)
        default_visible_users = getattr(self, 'visible_users', [])
        
        try: self._delProperty('statuses')
        except ValueError: pass
        try: self._delProperty('auto_open_status')
        except ValueError: pass
        try: self._delProperty('initial_status')
        except ValueError: pass
        try: self._delProperty('categories')
        except ValueError: pass
        try: self._delProperty('groups')
        except ValueError: pass
        try: self._delProperty('max_priority')
        except ValueError: pass
        try: self._delProperty('max_display_num')
        except ValueError: pass
        try: self._delProperty('suppress_statuses')
        except ValueError: pass
        try: self._delProperty('global_observers')
        except ValueError: pass
        try: self._delProperty('email_from_addr')
        except ValueError: pass 
        try: self._delProperty('mailhost_id')
        except ValueError: pass
        try: self._delProperty('visible_users')
        except ValueError: pass
        
        self.manage_addProperty('statuses', default_statuses, 'lines')
        self.manage_addProperty('auto_open_status',
                                 default_auto_open_status, 'string')
        self.manage_addProperty('initial_status',
                                 default_initial_status, 'string')
        self.manage_addProperty('categories', default_categories, 'lines')
        self.manage_addProperty('groups', default_groups, 'lines')
        self.manage_addProperty('max_priority',
                                 default_max_priority, 'int')
        self.manage_addProperty('max_display_num',
                                 default_max_display_num, 'int')
        self.manage_addProperty('suppress_statuses', default_suppress_statuses,
                                'lines')
        self.manage_addProperty('global_observers', default_global_observers,
                                'lines')
        self.manage_addProperty('email_from_addr', default_email_from_addr,
                                'string')
        self.manage_addProperty('mailhost_id', default_mailhost_id,
                                'string')
        self.manage_addProperty('visible_users', default_visible_users,
                                'lines')

    security.declareProtected('View management screens', 'update_user_folder')
    def update_user_folder(self):
        self.manage_addFolder(self.user_folder_name, 'zshrink users')
        
    security.declarePrivate('__initialise_product')
    def __initialise_product(self, file_attachment_path):
        if file_attachment_path:
            ob = LocalFS('file_attachments', 'File Attachments Repository',
                         file_attachment_path, None, None)
            self._setObject('file_attachments', ob)
            
        self.update_user_folder()
        self.update_catalog()
        self.update_properties()
        self.update_user_interface()
        
    security.declareProtected('View management screens', 'manage_addIssueTrack')
    def manage_addIssueTrack(self, id, file_attachment_path, RESPONSE=None):
        """ Add a IssueTrack object to a folder.
        
        """
        obj = IssueTrack(id)
        self._setObject(id, obj)
        
        issuetrack = getattr(self.aq_parent.this(), id)
        
        issuetrack.post_install(file_attachment_path)
        
        if RESPONSE:
            RESPONSE.redirect('%s/index_html' % id)
            
    security.declareProtected('View management screens',
                              'manage_addIssueTrackForm')
    def manage_addIssueTrackForm(self):
        """ The form used to get the instance id from the user.
        
        """
        return '''
        <html><body bgcolor="#FFFFFF">
        Please type the id and base directory of the ZShrink instance:<br>
        <form name="form" action="manage_addIssueTrack" method="post"><br>
        id<br>
        <input type="text" name="id"><br>
        file attachment storage directory<br>
        <input type="text" name="file_attachment_path"><br>
        <input type="submit" value="add">
        </form>
        </body>
        </html>
        '''
    
    #
    # Live fixes to previous ZShrink versions
    ############################################################################
    security.declareProtected('View management screens', 'apply_fix')
    def apply_fix(self):
        """ Apply new 'fixes' to an instance of ZShrink

        """
        version = getattr(self, '_version', 0)
                
        if version < 0.32:
            self.manage_addProperty('suppress_statuses',
                                    self.default_suppress_statuses, 'lines')

        if version < 0.35:
            self.update_user_folder()

            for key in self.issues.keys():
                self.issues[key].observers = []
                self.issues[key].last_notification_time = 0

            # add user objects to the user folder
            users_folder = getattr(self, self.user_folder_name)
            visible_users = [test.strip() for test in self.visible_users]
            for user in self.visible_users:
                user = user.strip()
                userObj = User(user)
                users_folder._setObject(user, userObj, user)
                
        self.update_user_interface()
        self.update_properties()
        self.update_catalog()
        
        self._version = self.version

        self._p = 1
        
        return 'ZShrink instance updated to version %s from version %s' % \
               (self._version, version)
    
###############################################################################
                                                                          #EOF#
                                                                          #####
